
__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of string operations. "
)

__all__ = """
    str_len
    str_slice
    find_all
    sorted_dict_repr
    enclosed_object
    tokenize
""".split()

def str_len(str_:str, r:int=2):
    """
    Returen the ASCII string length of `str_`. 
    
    Args:
        r: bytes a wide character stands for. 
    
    Example:
    ----------
    >>> print(str_len("12"), len("你好"), str_len("你好"))
    2 2 4
    """
    return int(len(str_) + (r-1) * len([c for c in str_ if ord(u'\u4e00') <= ord(c) <= ord(u'\u9fa5')]))

def find_all(str_:str, key:str):
    """
    Returen all the starting indices of string `key` in string `str_`. 
    
    Example:
    ----------
    >>> find_all("abcaa", 'a')
    [0, 3, 4]
    """
    p, indices = -1, []
    while True:
        p = str_.find(key, p + 1)
        if p < 0: break
        indices.append(p)
    return indices

def str_slice(str_:str, indices:list):
    """
    Split the string `str_` by breaks in list `indices`.
    
    Example:
    ----------
    >>> str_slice("abcaa", [2,4])
    ["ab", "ca", "a"]
    """
    indices.insert(0, 0); indices.append(len(str_))
    return [str_[indices[i]:indices[i+1]] for i in range(len(indices) - 1)]

def sorted_dict_repr(d:dict, order:list):
    """
    Representer of dictionary `d`, with key order `order`.
    
    Example:
    ----------
    >>> sorted_dict_repr({'a': 1, '0': 3}, ['0', 'a'])
    "{'0': 3, 'a': 1}"
    """
    return '{' + ', '.join([f"{repr(k)}: {repr(d[k])}" for k in order]) + '}'

def enclosed_object(str_, by=["([{", ")]}", "$`'\""], start=0):
    """
    Return the first object enclosed with a whole pair of parenthesis in `str_` after index `start`.
    
    Example:
    ----------
    >>> enclosed_object("function(something inside), something else. ")
    function(something inside)
    """
    if len(by) == 3: left, right, both = by
    elif len(by) == 2: left, right = by; both = ""
    elif len(by) == 1: left = ""; right = ""; both = by[0]
    else: raise TypeError("Invalid argument `by` for function `tokenize`. ")
    depth = {'all': 0}
    for i in range(start, len(str_)):
        s = str_[i]
        if s in right:
            if depth.get(s, 0) == 0: return str_[start:i]
            assert depth[s] > 0 and depth['all'] > 0
            depth[s] -= 1
            depth['all'] -= 1
            if depth[s] == 0 and depth['all'] == 0: return str_[start:i+1]
        elif s in left:
            r = right[left.index(s)]
            depth.setdefault(r, 0)
            depth[r] += 1
            depth['all'] += 1
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] > 0:
                depth[s] -= 1;
                depth['all'] -= 1
                if depth[s] == 0 and depth['all'] == 0: return str_[start:i+1]
            else: depth[s] += 1; depth['all'] += 1
    raise RuntimeError(f"Cannot find enclosed object from string {repr(str_)}.")

def tokenize(str_:str, sep=[' ', '\n'], by=["([{", ")]}", "$`'\""], start=0, strip='', keep_empty=True):
    """
    Split the string `str_` by elements in `sep`, but keep enclosed objects not split.
    
    Example:
    ----------
    >>> tokenize("function(something inside), something else. ")
    ["function(something inside),", "something", "else.", ""]
    """
    if isinstance(sep, str): sep = [sep]
    if len(by) == 3: left, right, both = by
    elif len(by) == 2: left, right = by; both = ""
    elif len(by) == 1: left = ""; right = ""; both = by[0]
    else: raise TypeError("Invalid argument `by` for function `tokenize`. ")
    depth = {'all': 0}
    tokens = []
    p = start
    for i in range(start, len(str_)):
        s = str_[i]
        both_done = False
        if s in right:
            if depth.get(s, 0) == 0: break
            assert depth[s] > 0 and depth['all'] > 0
            depth[s] -= 1
            depth['all'] -= 1
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] > 0:
                depth[s] -= 1
                depth['all'] -= 1
                both_done = True
        if depth['all'] == 0:
            for x in sep:
                if str_[i:i + len(x)] == x:
                    t = str_[p:i].strip(strip)
                    if keep_empty or t != '': 
                        tokens.append(t)
                    p = i + len(x)
        if s in left:
            r = right[left.index(s)]
            depth.setdefault(r, 0)
            depth[r] += 1
            depth['all'] += 1
        elif both_done: pass
        elif s in both and str_[i-1] != '\\':
            depth.setdefault(s, 0)
            if depth[s] == 0:
                depth[s] += 1
                depth['all'] += 1
    t = str_[p:].strip(strip)
    if keep_empty or t != '': 
        tokens.append(t)
    return tokens




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to handle the outputs.",
    help = "Use `sprint = SPrint()` to start printing to a string, etc. "
)

__all__ = """
    no_print
    str_print
    SPrint
""".split()

import sys

class _strIO:
    def __init__(self): self._str_ = ''
    def write(self, s): self._str_ += s
    def __str__(self): return self._str_
    def split(self, c=None): return self._str_.split(c)

class NPrint:
    """
    Suppress the outputs, designed for instance `no_print`.
    """
    def __enter__(self):
        self.io = _strIO()
        self.old_io = sys.stdout
        sys.stdout = self.io
        return self.io
    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout = self.old_io

no_print = NPrint()
"""
Suppress the outputs.

Example:
----------
>>> print("something in the front")
>>> with no_print:
>>>     print("something in the middle")
>>> print("something in behind")
something in the front
something in behind
"""

class SPrint:
    """
    Print to a string.
    
    Inputs:
        init_text[str]: initial text, as the head of the resulting string.
        sep[str]: the seperator between different elements, just like the built-in function `print`.
        end[str]: the ending string of each output, just like the built-in function `print`.

    Example:
    ----------
    >>> output = SPrint("!>> ")
    >>> output("Use it", "like", 'the function', "'print'.", sep=' ')
    !>> Use it like the function 'print'.
    >>> output("A return is added automatically each time", end=".")
    !>> Use it like the function 'print'.
    A return is added automatically each time.
    >>> output.text
    !>> Use it like the function 'print'.
    A return is added automatically each time.
    """

    def __init__(self, init_text='', sep=' ', end='\n'):
        self.text = init_text
        self.def_sep = sep
        self.def_end = end

    def __call__(self, *parts, sep=None, end=None):
        if not sep: sep = self.def_sep
        if not end: end = self.def_end
        self.text += sep.join([str(x) for x in parts if str(x)]) + end
        return self.text

    def __str__(self): return self.text

    def clear(self): self.text = ""
    
    def save(self, file_path):
        with open(file_path, 'w') as fp:
            fp.write(self.text)
            
    def append_to(self, file_path):
        with open(file_path, 'a') as fp:
            fp.write(self.text)

str_print = SPrint()
"""
Print to a string.

Example:
----------
>>> str_print.clear()
>>> str_print("Use it", "like", 'the function', "'print'.", sep=' ')
Use it like the function 'print'.
>>> str_print("A return is added automatically each time", end=".")
Use it like the function 'print'.
A return is added automatically each time.
>>> str_print.text
Use it like the function 'print'.
A return is added automatically each time.
"""




__info__ = dict(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to manage the environment.",
    help = "Use `get_**` to obtain the the variables etc. outside the function. "
)

__all__ = """
    get_environ_locals
    get_environ_globals
    get_args_expression
""".split()

import sys

def _mid(x): return x[1] if len(x) > 1 else x[0]
def _rawname(s): return _mid(str(s).split("'"))

stack_error = lambda x: TypeError(f"Unexpected function stack for {x}, please contact the developer for further information. ")

def _get_frames():
    frames = []
    frame = sys._getframe()
    fname = frame.f_back.f_code.co_name
    while frame is not None:
        frame_file = _rawname(frame)
        if frame_file.startswith('<') and frame_file.endswith('>') and frame_file != '<stdin>':
            frame = frame.f_back
            continue
        frames.append(frame)
        if len(frames) >= 4: return frames[2:]
        frame = frame.f_back
    raise stack_error(fname)

def get_environ_locals():
    _, client_frame = _get_frames()
    return client_frame.f_locals

def get_environ_globals():
    _, client_frame = _get_frames()
    return client_frame.f_globals
    
def get_args_expression():
    module_frame, client_frame = _get_frames()
    func_name = module_frame.f_code.co_name
    with open(client_frame.f_code.co_filename) as fp:
        for _ in range(client_frame.f_lineno-1): fp.readline()
        l = fp.readline()
        if func_name not in l: raise stack_error
        return l.split(func_name)[-1].split(';')[0].strip().strip('()')




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File of list operations. "
)

__all__ = """
    prod
    argmin
    argmax
    flat_list
    item
    to_list
    to_tuple
""".split()

from pycamia import avouch, touch

def prod(*x):
    """
    Returns the product of elements, just like built-in function `sum`.
    
    Example:
    ----------
    >>> prod([5, 2, 1, 4, 2])
    80
    """
    if len(x) == 1 and isinstance(x[0], list): x = x[0]
    p = 1
    for i in x: 
        if hasattr(i, "__mul__"):
            p *= i
    return p

def argmin(y, x=None):
    """
    Find the indices of minimal element in `y` given domain `x`.
    
    Example:
    ----------
    >>> argmin([0, 2, 1, 4, 2], [1, 3, 4])
    [1, 4]
    """
    if x is None: x = range(len(y))
    if len(x) <= 0: return []
    m = min([y[i] for i in x])
    return [i for i in x if y[i] == m]

def argmax(y, x):
    """
    Find the indices of maximal element in `y` given domain `x`.
    
    Example:
    ----------
    >>> argmin([0, 2, 1, 4, 2], [1, 3, 4])
    [3]
    """
    if x is None: x = range(len(y))
    if len(x) <= 0: return []
    m = max([y[i] for i in x])
    return [i for i in x if y[i] == m]

def flat_list(list_):
    """
    Flat the nested lists `list_`.
    
    Example:
    ----------
    >>> flat_list([0, 2, [1, 4, 2], [1, 3, 4]])
    [0, 2, 1, 4, 2, 1, 3, 4]
    """
    # Deprecated realization of the function, as elements may be strings with characters '[' or ']'.
    # items = str(list_).replace('[', '').replace(']', '').split(',')
    # return list(eval(','.join([x for x in items if x.strip() != ''])))
    flattened = []
    for x in list_:
        if isinstance(x, list):
            flattened.extend(flat_list(x))
        else: flattened.append(x)
    return flattened

def item(list_):
    """
    Assert if the length of the list `list_` is not 1 and return the only element. 
    
    Example:
    ----------
    >>> item([0])
    0
    >>> item([1,2])
    AssertError: ...
    """
    avouch(len(list_) == 1, f"Failure in itemize as the length of the list {repr(list_)} is not 1. ")
    return list_[0]

def to_list(x):
    """
    Try to cast element `x` into a list
    
    Example:
    ----------
    >>> item(0)
    [0]
    >>> to_list((1,2))
    [1,2]
    """
    func_candidates = ['tolist', 'to_list', 'aslist', 'as_list', '__list__']
    for fname in func_candidates:
        if hasattr(x, fname) and callable(getattr(x, fname)): return getattr(x, fname)()
    return touch(lambda: list(x), [x])

def to_tuple(x):
    """
    Try to cast element `x` into a tuple
    
    Example:
    ----------
    >>> item(0)
    (0,)
    >>> to_tuple([1,2])
    (1,2)
    """
    func_candidates = ['totuple', 'to_tuple', 'astuple', 'as_tuple', '__tuple__']
    for fname in func_candidates:
        if hasattr(x, fname) and callable(getattr(x, fname)): return getattr(x, fname)()
    return touch(lambda: tuple(x), (x,))




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to record time."
)

__all__ = """
    time_this
    Timer
    Jump
    scope
    jump
    Workflow
    periodic
""".split()

with __info__:
    import time
    from functools import wraps
    from threading import Timer

def time_this(func):
    """
    A function wrapper of function `func` that outputs the time used to run it. 

    Example:
    ----------
    @timethis
    >>> def func_to_run(*args):
    ...     # inside codes
    ... 
    >>> func_to_run(*input_args)
    # some outputs
    [func_to_run takes 0.001s]
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        if hasattr(getattr(func, '__wrapped__', func), '__name__'):
            print("[%s takes %lfs]"%(func.__name__, end-start))
        else:
            print("[%s takes %lfs]"%(func.__class__.__name__, end-start))
        return result
    return wrapper

class Timer(object):
    """
    An environment that outputs the time used to run codes within. 

    Example:
    ----------
    >>> with Timer("test"):
    ...     # inside codes
    ... 
    # some outputs
    [test takes 0.001s]
    """
    def __init__(self, name='', timing=True):
        if not timing: name = ''
        self.name = name
        self.nround = 0
    def __enter__(self):
        self.start = time.time()
        self.prevtime = self.start
        return self
    def round(self, name = ''):
        self.nround += 1
        self.end = time.time()
        if self.name:
            if not name: name = "%s(round%d)"%(self.name, self.nround)
            print("[%s takes %lfs]"%(name, self.end - self.prevtime))
        self.prevtime = self.end
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type == RuntimeError and str(exc_value) == "JUMP": return True
        if self.name:
            print("[%s%s takes %lfs]"%
                  (self.name, '' if self.nround == 0 else "(all)", time.time() - self.start))

class Jump(object):
    """
    Creates a Jump RuntimeError, designed for instance `jump`. 
    """
    def __init__(self, jump=None): self.jump = True if jump is None else jump
    def __enter__(self):
        def dojump(): raise RuntimeError("JUMP")
        if self.jump: dojump()
        else: return dojump
    def __exit__(self, *args): pass
    def __call__(self, condition): return Jump(condition)
    
def scope(name, timing=True):
    """
    An allias of timer to better organize the codes. 
    
    Inputs:
        name (str): the name of the scope, used to display. 
        timing (bool): whether to show the time span or not. 

    Example:
    ----------
    >>> with scope("test"):
    ...     # inside codes
    ... 
    # some outputs
    [scope test takes 0.001s]
    """
    return Timer("scope " + name, timing)

jump = Jump()
"""
The jumper, one can use it along with `scope`(or `Timer`) to jump a chunk of codes. 

Example:
----------
>>> with scope("test"), jump:
...     # inside codes
... 
# nothing, the inside codes do not run
>>> with scope("test"), jump as stop:
...     print('a')
...     stop()
...     print('b')
... 
a
"""

class Workflow:
    """
    A structure to create a series of workflow. 
    
    Note:
        Remember to manually add `, {workflow_name}.jump` after `with` so that 
        we can control it. See the example. 
    
    Args:
        *args: the list of scope names to run. 

    Example:
    ----------
    >>> run = Workflow("read data", "run method", "visualization")
    ... with run("read data"), run.jump:
    ...     print(1, end='')
    ... with run("pre-processing"), run.jump:
    ...     print(2, end='')
    ... with run("run method"), run.jump:
    ...     print(3, end='')
    ... with run("visualization"), run.jump:
    ...     print(4, end='')
    ... 
    1[read data takes 0.000022s]
    3[run method takes 0.000008s]
    4[visualization takes 0.000006s]
    """
    def __init__(self, *args): self.workflow = args
    def __call__(self, key): self.key=key; return Timer(key)
    def __getattr__(self, k): return self(k)
    def __getitem__(self, k): return self(k)
    @property
    def j(self): return Jump(self.key not in self.workflow)
    @property
    def jump(self): return Jump(self.key not in self.workflow)

class TimerCtrl(Timer):
    """
    Creates a Time Handler, designed for function `periodic`. 
    """

    def __init__(self, seconds, function):
        Timer.__init__(self, seconds, function)
        self.isCanceled = False
        self.seconds = seconds
        self.function = function
        self.funcname = function.__name__
        self.startTime = time.time()

    def cancel(self):
        Timer.cancel(self)
        self.isCanceled = True

    def is_canceled(self): return self.isCanceled

    def setFunctionName(self, funcname): self.funcname = funcname

    def __str__(self):
        return "%5.3fs to run next "%(self.seconds + self.startTime -
                                      time.time()) + self.funcname


def periodic(period, maxiter=float('Inf')):
    """
    A function wrapper to repeatedly run the wrapped function `period`.
    
    Args:
        maxiter (int): the number of iterations. 

    Example:
    ----------
    >>> i = 1
    ... @periodic(1)
    ... def func():
    ...     print(i)
    ...     i+= 1
    ... 
    1
    2
    3
    [Output every 1s, and GO ON...]
    """
    def wrap(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            global count, timer_ctrl
            try:
                if timer_ctrl.is_canceled(): return
            except NameError: pass
            timer_ctrl = TimerCtrl(period, lambda : wrapper(*args, **kwargs))
            timer_ctrl.setFunctionName(func.__name__)
            timer_ctrl.start()
            ret = func(timer_ctrl, *args, **kwargs)
            try: count += 1
            except Exception: count = 1
            if count >= maxiter: return
            return ret
        return wrapper
    return wrap
